# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
import re
from pathlib import Path

FIXTURES = Path(__file__).resolve().parent / "data"

TWO_SPACES_REGEX = re.compile(r" {2}")


def change_split_content(
    load_entities: bool,
    transcription_entities_worker_version: str | bool,
    keep_spaces: bool,
    split_content: dict,
    tokens: list,
    expected_labels: dict = [],
):
    # Transcriptions with worker version are in lowercase
    if transcription_entities_worker_version:
        for split in split_content:
            for element_id in split_content[split]:
                split_content[split][element_id]["text"] = split_content[split][
                    element_id
                ]["text"].lower()
        for split in expected_labels:
            for image in expected_labels[split]:
                expected_labels[split][image] = expected_labels[split][image].lower()

    # If we do not load entities, remove tokens
    if not load_entities:
        token_translations = {ord(token): None for token in tokens}
        for split in split_content:
            for element_id in split_content[split]:
                split_content[split][element_id]["text"] = split_content[split][
                    element_id
                ]["text"].translate(token_translations)
        for split in expected_labels:
            for image in expected_labels[split]:
                expected_labels[split][image] = expected_labels[split][image].translate(
                    token_translations
                )

    # Replace double spaces with regular space
    if not keep_spaces:
        for split in split_content:
            for element_id in split_content[split]:
                split_content[split][element_id]["text"] = TWO_SPACES_REGEX.sub(
                    " ", split_content[split][element_id]["text"]
                )
        for split in expected_labels:
            for image in expected_labels[split]:
                expected_labels[split][image] = TWO_SPACES_REGEX.sub(
                    " ", expected_labels[split][image]
                )

    return split_content, expected_labels



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
"""
Predict on an image using a trained DAN model.
"""

from pathlib import Path

from dan.ocr.predict.attention import Level
from dan.ocr.predict.inference import run
from dan.utils import parse_tokens


def add_predict_parser(subcommands) -> None:
    parser = subcommands.add_parser(
        "predict",
        description=__doc__,
        help=__doc__,
    )
    # Required arguments.
    parser.add_argument(
        "--image-dir",
        type=Path,
        help="Path to the folder where the images to predict are stored.",
    )
    parser.add_argument(
        "--model",
        type=Path,
        help="Path to the directory containing the model, the YAML parameters file and the charset file to use for prediction.",
        required=True,
    )
    parser.add_argument(
        "--font",
        type=Path,
        default=Path("fonts/LinuxLibertine.ttf"),
        help="Path to the font file to use for the GIF of the attention map.",
    )
    parser.add_argument(
        "--maximum-font-size",
        type=int,
        default=32,
        help="Maximum font size to use for the GIF of the attention map.",
    )
    parser.add_argument(
        "--output",
        type=Path,
        help="Path to the output folder.",
        required=True,
    )
    # Optional arguments.
    parser.add_argument(
        "--tokens",
        type=parse_tokens,
        required=False,
        help="Path to a yaml file containing a mapping between starting tokens and end tokens. Needed for entities.",
    )
    parser.add_argument(
        "--image-extension",
        type=str,
        help="The extension of the images in the folder.",
        default=".jpg",
    )
    parser.add_argument(
        "--temperature",
        type=float,
        default=1.0,
        help="Temperature scaling scalar parameter.",
        required=False,
    )
    parser.add_argument(
        "--confidence-score",
        action="store_true",
        help="Whether to return confidence scores.",
        required=False,
    )
    parser.add_argument(
        "--confidence-score-levels",
        default=[],
        type=Level,
        nargs="+",
        help=f"Levels of confidence scores. Should be a list of any combinaison of {list(map(str, Level))}.",
        required=False,
    )
    parser.add_argument(
        "--attention-map",
        action="store_true",
        help="Whether to plot attention maps.",
        required=False,
    )
    parser.add_argument(
        "--attention-map-level",
        type=Level,
        default=Level.Line,
        help=f"Level to plot the attention maps. Should be in {list(map(str, Level))}.",
        required=False,
    )
    parser.add_argument(
        "--attention-map-scale",
        type=float,
        default=0.5,
        help="Image scaling factor before creating the GIF.",
        required=False,
    )
    parser.add_argument(
        "--alpha-factor",
        help="""Alpha factor that controls how much the attention map is shown
                    to the user during prediction. (typically between .5 and 1)
                    (1 is no change, lower shows more of the attention)""",
        type=float,
        default=0.9,
        required=False,
    )
    parser.add_argument(
        "--color-map",
        help="A matplotlib colormap to use for the attention maps.",
        type=str,
        default="nipy_spectral",
        required=False,
    )
    parser.add_argument(
        "--word-separators",
        default=[" ", "\n"],
        type=str,
        nargs="+",
        help="String separators used to split text into words.",
        required=False,
    )
    parser.add_argument(
        "--line-separators",
        default=["\n"],
        type=str,
        nargs="+",
        help="String separators used to split text into lines.",
        required=False,
    )
    parser.add_argument(
        "--predict-objects",
        action="store_true",
        help="Whether to output objects when plotting attention maps.",
        required=False,
    )
    parser.add_argument(
        "--max-object-height",
        help="Maximum height for predicted objects. If set, grid search segmentation will be applied and width will be normalized to element width.",
        type=int,
        default=None,
    )
    parser.add_argument(
        "--gpu-device",
        help="Use a specific GPU if available.",
        type=int,
        required=False,
    )
    parser.add_argument(
        "--batch-size",
        help="Size of prediction batches.",
        type=int,
        default=1,
        required=False,
    )
    parser.add_argument(
        "--start-token",
        help="Use a specific starting token at the beginning of the prediction. Useful when making predictions on different single pages.",
        type=str,
        required=False,
    )
    parser.add_argument(
        "--use-language-model",
        help="Whether to use an explicit language model to rescore text hypotheses.",
        action="store_true",
        required=False,
    )
    parser.add_argument(
        "--compile-model",
        help="Whether to compile the model. Recommended to speed up inference.",
        action="store_true",
        required=False,
    )
    parser.add_argument(
        "--dynamic-mode",
        help="Whether to use the dynamic mode during model compilation. Recommended for prediction on images of variable size.",
        action="store_true",
        required=False,
    )
    parser.set_defaults(func=run)



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
import json
import uuid
from operator import itemgetter
from typing import List

import pytest

from arkindex_export import (
    Dataset,
    DatasetElement,
    Element,
    ElementPath,
    Entity,
    EntityType,
    Image,
    ImageServer,
    Transcription,
    TranscriptionEntity,
    WorkerRun,
    WorkerVersion,
    database,
)
from dan.datasets.extract.arkindex import TEST_NAME, TRAIN_NAME, VAL_NAME
from tests import FIXTURES


@pytest.fixture()
def mock_database(tmp_path_factory):
    def create_transcription_entity(
        transcription: Transcription,
        worker_version: str | None,
        worker_run: str | None,
        type: str,
        name: str,
        offset: int,
    ) -> None:
        entity_type, _ = EntityType.get_or_create(
            name=type, defaults={"id": f"{type}_id"}
        )
        entity = Entity.create(
            id=str(uuid.uuid4()),
            name=name,
            type=entity_type,
            worker_version=worker_version,
            worker_run=worker_run,
        )
        TranscriptionEntity.create(
            id=str(uuid.uuid4()),
            entity=entity,
            length=len(name),
            offset=offset,
            transcription=transcription,
            worker_version=worker_version,
            worker_run=worker_run,
        )

    def create_transcriptions(element: Element, entities: List[dict]) -> None:
        if not entities:
            return

        # Add transcription with entities
        entities = sorted(entities, key=itemgetter("offset"))

        # We will add extra spaces to test the "keep_spaces" parameters of the "extract" command
        for offset, entity in enumerate(entities[1:], start=1):
            entity["offset"] += offset

        for worker_version, worker_run in [
            (None, None),
            ("worker_version_id", "worker_run_id"),
        ]:
            transcription_suffix = ""
            # Use different transcriptions to filter by worker version
            if worker_version and worker_run:
                transcription_suffix = "source"
                for entity in entities:
                    entity["name"] = entity["name"].lower()

            transcription = Transcription.create(
                id=element.id + transcription_suffix,
                # Add extra spaces to test the "keep_spaces" parameters of the "extract" command
                text="  ".join(map(itemgetter("name"), entities)),
                element=element,
                worker_version=worker_version,
                worker_run=worker_run,
            )

            for entity in entities:
                create_transcription_entity(
                    transcription=transcription,
                    worker_version=worker_version,
                    worker_run=worker_run,
                    **entity,
                )

    def create_element(id: str, parent: Element | None = None) -> None:
        element_path = (FIXTURES / "extraction" / "elements" / id).with_suffix(".json")
        element_json = json.loads(element_path.read_text())

        element_type = element_json["type"]
        image_path = (FIXTURES / "extraction" / "images" / id).with_suffix(".jpg")

        polygon = element_json.get("polygon")
        # Always use page images because polygons are based on the full image
        image, _ = (
            Image.get_or_create(
                id=id + "-image",
                defaults={
                    "server": image_server,
                    # Use path to image instead of actual URL since we won't be doing any download
                    "url": image_path,
                    "width": 0,
                    "height": 0,
                },
            )
            if polygon
            else (None, False)
        )

        element = Element.create(
            id=id,
            name=id,
            type=element_type,
            image=image,
            polygon=json.dumps(polygon) if polygon else None,
            created=0.0,
            updated=0.0,
        )

        if parent:
            ElementPath.create(id=str(uuid.uuid4()), parent=parent, child=element)

        create_transcriptions(
            element=element,
            entities=element_json.get("transcription_entities", []),
        )

        # Recursive function to create children
        for child in element_json.get("children", []):
            create_element(id=child, parent=element)

    MODELS = [
        WorkerVersion,
        WorkerRun,
        ImageServer,
        Image,
        Dataset,
        DatasetElement,
        Element,
        ElementPath,
        EntityType,
        Entity,
        Transcription,
        TranscriptionEntity,
    ]

    # Initialisation
    tmp_path = tmp_path_factory.mktemp("data")
    database_path = tmp_path / "db.sqlite"
    database.init(
        database_path,
        pragmas={
            # Recommended settings from peewee
            # http://docs.peewee-orm.com/en/latest/peewee/database.html#recommended-settings
            # Do not set journal mode to WAL as it writes in the database
            "cache_size": -1 * 64000,  # 64MB
            "foreign_keys": 1,
            "ignore_check_constraints": 0,
            "synchronous": 0,
        },
    )
    database.connect()

    # Create tables
    database.create_tables(MODELS)

    image_server = ImageServer.create(
        url="http://image/server/url",
        display_name="Image server",
    )

    WorkerVersion.create(
        id="worker_version_id",
        slug="worker_version",
        name="Worker version",
        repository_url="http://repository/url",
        revision="main",
        type="worker",
    )

    WorkerRun.create(
        id="worker_run_id",
        worker_version="worker_version_id",
    )

    # Create dataset
    split_names = [VAL_NAME, TEST_NAME, TRAIN_NAME]
    dataset = Dataset.create(
        id="dataset_id",
        name="Dataset",
        state="complete",
        sets=",".join(split_names),
        description="My Dataset",
    )

    # Create dataset elements
    for split in split_names:
        element_path = (FIXTURES / "extraction" / "elements" / split).with_suffix(
            ".json"
        )
        element_json = json.loads(element_path.read_text())

        # Recursive function to create children
        for child in element_json.get("children", []):
            create_element(id=child)

            # Linking the element to the dataset split
            DatasetElement.create(
                id=child, element_id=child, dataset=dataset, set_name=split
            )

    # Create data for entities extraction tests
    # Create transcription
    transcription = Transcription.create(
        id="tr-with-entities",
        text="The great king Charles III has eaten \nwith us.",
        element=Element.select().first(),
    )

    # Create worker version and worker run
    for nestation in ("nested", "non-nested", "special-chars"):
        WorkerVersion.create(
            id=f"worker-version-{nestation}-id",
            slug=nestation,
            name=nestation,
            repository_url="http://repository/url",
            revision="main",
            type="worker",
        )
        WorkerRun.create(
            id=f"worker-run-{nestation}-id",
            worker_version=f"worker-version-{nestation}-id",
        )

    # Create entities
    for entity in [
        # Non-nested entities
        {
            "source": "non-nested",
            "type": "adj",
            "name": "great",
            "offset": 4,
        },
        {
            "source": "non-nested",
            "type": "name",
            "name": "Charles",
            "offset": 15,
        },
        {
            "source": "non-nested",
            "type": "person",
            "name": "us",
            "offset": 43,
        },
        # Nested entities
        {
            "source": "nested",
            "type": "fullname",
            "name": "Charles III",
            "offset": 15,
        },
        {
            "source": "nested",
            "type": "name",
            "name": "Charles",
            "offset": 15,
        },
        {
            "source": "nested",
            "type": "person",
            "name": "us",
            "offset": 43,
        },
        # Special characters
        {
            "source": "special-chars",
            "type": "Arkindex's entity",
            "name": "great",
            "offset": 4,
        },
        {
            "source": "special-chars",
            "type": '"Name" (1)',
            "name": "Charles",
            "offset": 15,
        },
        {
            "source": "special-chars",
            "type": "Person /!\\",
            "name": "us",
            "offset": 43,
        },
    ]:
        source = entity.pop("source")
        create_transcription_entity(
            transcription=transcription,
            worker_version=f"worker-version-{source}-id",
            worker_run=f"worker-run-{source}-id",
            **entity,
        )

    return database_path


@pytest.fixture
def training_config():
    return json.loads((FIXTURES.parent.parent / "configs" / "tests.json").read_text())


@pytest.fixture
def evaluate_config():
    return json.loads((FIXTURES.parent.parent / "configs" / "eval.json").read_text())


@pytest.fixture
def split_content():
    splits = json.loads((FIXTURES / "extraction" / "split.json").read_text())
    for split in splits:
        for element_id in splits[split]:
            splits[split][element_id]["image"]["iiif_url"] = splits[split][element_id][
                "image"
            ]["iiif_url"].replace("{FIXTURES}", str(FIXTURES))

    return splits



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
import logging
from io import BytesIO
from typing import List

import requests
from PIL import Image, ImageOps
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

logger = logging.getLogger(__name__)

# See http://docs.python-requests.org/en/master/user/advanced/#timeouts
DOWNLOAD_TIMEOUT = (30, 60)


def _retry_log(retry_state, *args, **kwargs):
    logger.warning(
        f"Request to {retry_state.args[0]} failed ({repr(retry_state.outcome.exception())}), "
        f"retrying in {retry_state.idle_for} seconds"
    )


@retry(
    stop=stop_after_attempt(3),
    wait=wait_exponential(multiplier=2),
    retry=retry_if_exception_type(requests.RequestException),
    before_sleep=_retry_log,
    reraise=True,
)
def _retried_request(url: str) -> requests.Response:
    resp = requests.get(url, timeout=DOWNLOAD_TIMEOUT)
    resp.raise_for_status()
    return resp


def download_image(url: str) -> Image.Image:
    """
    Download an image and open it with Pillow
    """
    assert url.startswith("http"), "Image URL must be HTTP(S)"

    # Download the image
    # Cannot use stream=True as urllib's responses do not support the seek(int) method,
    # which is explicitly required by Image.open on file-like objects
    try:
        resp = _retried_request(url)
    except requests.HTTPError as e:
        if "/full/" in url and 400 <= e.response.status_code < 500:
            # Retry with max instead of full as IIIF size
            resp = _retried_request(url.replace("/full/", "/max/"))
        else:
            raise e

    # Preprocess the image and prepare it for classification
    image = Image.open(BytesIO(resp.content)).convert("RGB")

    # Do not rotate JPEG images (see https://github.com/python-pillow/Pillow/issues/4703)
    image = ImageOps.exif_transpose(image)

    logger.debug(
        "Downloaded image {} - size={}x{}".format(url, image.size[0], image.size[1])
    )

    return image


def get_bbox(polygon: List[List[int]]) -> str:
    """
    Returns a comma-separated string of upper left-most pixel, width + height of the image
    """
    all_x, all_y = zip(*polygon)
    x, y = min(all_x), min(all_y)
    width, height = max(all_x) - x, max(all_y) - y
    return ",".join(list(map(str, [int(x), int(y), int(width), int(height)])))



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
from pathlib import Path
from typing import Iterable

import yaml

OFFSET = 86
LIMIT = 160

STARTING_TOKEN = "\u2460"


def get_token() -> Iterable[str]:
    offset = OFFSET

    while offset < LIMIT:
        yield chr(ord(STARTING_TOKEN) + offset % LIMIT)
        offset += 1

    raise Exception(f"More than {LIMIT} tokens asked")


def run(entities: Path, end_tokens: bool, output_file: Path) -> None:
    # Load extracted entities
    entities = yaml.safe_load(entities.read_text())

    # Generate associated starting/ending token
    token_generator = get_token()
    tokens = {}
    for entity in entities.get("entities", []):
        tokens[entity] = {
            "start": next(token_generator),
            "end": next(token_generator) if end_tokens else "",
        }

    # Save entities & tokens to YAML
    output_file.parent.mkdir(exist_ok=True, parents=True)
    output_file.write_text(
        yaml.safe_dump(tokens, explicit_start=True, allow_unicode=True, sort_keys=False)
    )



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
from dan.datasets.entities.extract import run
from tests import FIXTURES


def test_entities(mock_database, tmp_path):
    output_file = tmp_path / "entities.yml"

    run(database=mock_database, output_file=output_file)

    assert output_file.read_text() == (FIXTURES / "entities.yml").read_text()



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

"""
Convert DAN predictions to BIO format.
"""

import json
import logging
import re
from pathlib import Path
from typing import Dict, List

from tqdm import tqdm

from dan.utils import EntityType, parse_tokens

logger = logging.getLogger(__name__)


def add_convert_bio_parser(subcommands):
    parser = subcommands.add_parser(
        "convert",
        description=__doc__,
        help=__doc__,
    )
    parser.set_defaults(func=run)

    parser.add_argument(
        "predictions",
        type=Path,
        help="Path to a folder of DAN predictions.",
    )

    parser.add_argument(
        "--tokens",
        type=Path,
        help="Mapping between starting tokens and end tokens to extract text with their entities.",
        required=True,
    )

    parser.add_argument(
        "--output",
        type=Path,
        help="Where BIO files are saved. Will be created if missing.",
        required=True,
    )


def convert(text: str, ner_tokens: Dict[str, EntityType]) -> str:
    # Mapping to find a starting token for an ending token efficiently
    mapping_end_start: Dict[str, str] = {
        entity_type.end: entity_type.start for entity_type in ner_tokens.values()
    }
    # Mapping to find the entity name for a starting token efficiently
    mapping_start_name: Dict[str, str] = {
        entity_type.start: name for name, entity_type in ner_tokens.items()
    }

    starting_tokens: List[str] = mapping_start_name.keys()
    ending_tokens: List[str] = mapping_end_start.keys()
    has_ending_tokens: bool = set(ending_tokens) != {
        ""
    }  # Whether ending tokens are used

    # Spacing starting tokens and ending tokens (if necessary)
    tokens_spacing: re.Pattern = re.compile(
        r"([" + "".join([*starting_tokens, *ending_tokens]) + "])"
    )
    text: str = tokens_spacing.sub(r" \1 ", text)

    iob: List[str] = []  # List of IOB formatted strings
    entity_types: List[str] = []  # Encountered entity types
    inside: bool = False  # Whether we are inside an entity
    for token in text.split():
        # Encountering a starting token
        if token in starting_tokens:
            entity_types.append(token)

            # Stopping any current entity type
            inside = False

            continue

        # Encountering an ending token
        elif has_ending_tokens and token in ending_tokens:
            if not entity_types:
                logger.warning(
                    f"Missing starting token for ending token {token}, skipping the entity"
                )
                continue

            # Making sure this ending token closes the current entity
            assert (
                entity_types[-1] == mapping_end_start[token]
            ), f"Ending token {token} doesn't match the starting token {entity_types[-1]}"

            # Removing the current entity from the queue as it is its end
            entity_types.pop()

            # If there is still entities in the queue, we continue in the parent one
            # Else, we are not in any entity anymore
            inside = bool(entity_types)

            continue

        # The token is not part of an entity
        if not entity_types:
            iob.append(f"{token} O")
            continue

        # The token is part of at least one entity
        entity_name: str = mapping_start_name[entity_types[-1]]

        if inside:
            # Inside the same entity
            iob.append(f"{token} I-{entity_name}")
            continue

        # Starting a new entity
        iob.append(f"{token} B-{entity_name}")
        inside = True

    # Concatenating all formatted iob strings
    return "\n".join(iob)


def run(
    predictions: Path,
    tokens: Path,
    output: Path,
):
    # Create output folder
    output.mkdir(parents=True, exist_ok=True)

    # Load tokens
    ner_tokens = parse_tokens(tokens)

    for prediction in tqdm(
        list(predictions.glob("*.json")), desc="Converting predictions"
    ):
        data = json.loads(prediction.read_text())
        try:
            bio_representation = convert(data["text"], ner_tokens)
        except Exception as e:
            logger.error(f"Failed to convert {prediction.name}: {e}")
            continue

        (output / prediction.stem).with_suffix(".bio").write_text(bio_representation)



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
import json
import re
from argparse import ArgumentTypeError
from itertools import islice, takewhile
from operator import attrgetter
from pathlib import Path
from typing import Dict, NamedTuple

import torch
import torchvision.io as torchvision
import yaml


class MLflowNotInstalled(Exception):
    """
    Raised when MLflow logging was requested but the module was not installed
    """


class Token(NamedTuple):
    encoded: str
    display: str


class LMTokenMapping(NamedTuple):
    space: Token = Token("▁", " ")
    linebreak: Token = Token("↵", "\n")
    ctc: Token = Token("◌", "<ctc>")

    @property
    def display(self):
        return {a.encoded: a.display for a in self}

    @property
    def encode(self):
        return {a.display: a.encoded for a in self}

    def encode_token(self, token: str) -> str:
        return self.encode.get(token, token)


class EntityType(NamedTuple):
    start: str
    end: str = ""

    @property
    def offset(self):
        return len(self.start) + len(self.end)


def pad_sequences_1D(data, padding_value):
    """
    Pad data with padding_value to get same length
    """
    x_lengths = [len(x) for x in data]
    longest_x = max(x_lengths)
    padded_data = torch.ones((len(data), longest_x), dtype=torch.int32) * padding_value
    for i, x_len in enumerate(x_lengths):
        padded_data[i, :x_len] = torch.tensor(data[i][:x_len])
    return padded_data


def pad_images(images):
    """
    Pad the images so that they are at the top left of the large padded image.
    :param images: List of images as torch tensors.
    :return padded_images: A tensor containing all the padded images.
    """
    longest_x = max([x.shape[1] for x in images])
    longest_y = max([x.shape[2] for x in images])
    padded_images = torch.zeros((len(images), images[0].shape[0], longest_x, longest_y))
    for index, image in enumerate(images):
        padded_images[
            index,
            :,
            0 : image.shape[1],
            0 : image.shape[2],
        ] = image
    return padded_images


def read_image(path: str):
    """
    Read image with torch
    :param path: Path of the image to load.
    """
    img = torchvision.read_image(path, mode=torchvision.ImageReadMode.RGB)
    return img.to(dtype=torch.get_default_dtype()).div(255)


# Charset / labels conversion
def token_to_ind(labels, str):
    return [labels.index(c) for c in str]


def ind_to_token(labels, ind, oov_symbol=None):
    if oov_symbol is not None:
        res = []
        for i in ind:
            if i < len(labels):
                res.append(labels[i])
            else:
                res.append(oov_symbol)
    else:
        res = [labels[i] for i in ind]
    return "".join(res)


def fix_ddp_layers_names(model, to_ddp):
    """
    Rename the model layers if they were saved using DDP or if they will
    be used with DDP.
    :param model: Model to update.
    :param to_ddp: Convert layers names to be used by DDP.
    :return: The model with corrected layers names.
    """
    if to_ddp:
        return {
            ("module." + k if "module" not in k else k): v for k, v in model.items()
        }
    return {k.replace("module.", ""): v for k, v in model.items()}


def list_to_batches(iterable, n):
    "Batch data into tuples of length n. The last batch may be shorter."
    # list_to_batches('ABCDEFG', 3) --> ABC DEF G
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while batch := tuple(islice(it, n)):
        yield batch


def parse_tokens(filename: str) -> Dict[str, EntityType]:
    tokens = {
        name: EntityType(**tokens)
        for name, tokens in yaml.safe_load(Path(filename).read_text()).items()
    }
    end_tokens = map(attrgetter("end"), tokens.values())
    # Check that either
    if next(end_tokens):
        # - all entities have end token
        assert all(end_tokens), "Some entities have no end token"
    else:
        # - no entities have end tokens
        assert not any(end_tokens), "Some entities have end tokens"

    return tokens


def parse_tokens_pattern(tokens: list[EntityType]) -> re.Pattern[str]:
    starting_tokens = "".join(token.start for token in tokens)
    # Check if there are end tokens
    if all(token.end for token in tokens):
        # Matches a starting token, the corresponding ending token, and any text in between
        return re.compile(
            r"|".join(f"({token.start}.*?{token.end})" for token in tokens)
        )

    # Matches a starting token then any character that is not a starting token
    return re.compile(rf"([{starting_tokens}][^{starting_tokens}]*)")


def parse_charset_pattern(charset: list[str]) -> re.Pattern[str]:
    """
    Use (...) for tokens with a longer length than 1, otherwise, use [...].
    Longer words are matched first by the pattern.
    """

    tokens = sorted(charset, key=len)

    # 1 character length
    letters = list(takewhile(lambda t: len(t) == 1, tokens))

    pattern = ""

    if words := tokens[len(letters) :]:
        # More than 1 character length
        # (?:...)|(?:...)...
        # Create non capturing groups to be able to split with re.findall
        pattern += (
            r"|".join(f"(?:{re.escape(token)})" for token in reversed(words)) + "|"
        )

    # [...] used for 1-length tokens
    pattern += "[" + r"".join(map(re.escape, letters)) + "]"

    return re.compile(pattern)


def read_yaml(yaml_path: str) -> Dict:
    """
    Read YAML tokens file.
    :param yaml_path: Path of the YAML file to read.
    :return: The content of the read file.
    """
    filename = Path(yaml_path)
    assert filename.exists(), f"{yaml_path} does not resolve."
    try:
        return yaml.safe_load(filename.read_text())
    except yaml.YAMLError as e:
        raise ArgumentTypeError(e)


def read_json(json_path: str) -> Dict:
    """
    Read labels JSON file
    :param json_path: Path of the JSON file to read.
    :return: The content of the read file.
    """
    filename = Path(json_path)
    assert filename.exists(), f"{json_path} does not resolve."
    try:
        return json.loads(filename.read_text())
    except json.JSONDecodeError as e:
        raise ArgumentTypeError(e)


def read_txt(txt_path: str) -> str:
    """
    Read TXT file.
    :param txt_path: Path of the text file to read.
    :return: The content of the read file.
    """
    filename = Path(txt_path)
    assert filename.exists(), f"{txt_path} does not resolve."
    return filename.read_text()



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
import numpy as np
from torch.nn import Dropout, Dropout2d


class DropoutScheduler:
    def __init__(self, models, T=5e4):
        """
        T: number of gradient updates to converge
        """

        self.teta_list = list()
        self.init_teta_list(models)
        self.T = T
        self.step_num = 0

    def step(self, num):
        self.step_num += num

    def resume(self, step_num):
        self.step_num = step_num

    def init_teta_list(self, models):
        for model_name in models:
            self.init_teta_list_module(models[model_name])

    def init_teta_list_module(self, module):
        for child in module.children():
            if isinstance(child, Dropout) or isinstance(child, Dropout2d):
                self.teta_list.append([child, child.p])
            else:
                self.init_teta_list_module(child)

    def update_dropout_rate(self):
        for module, p in self.teta_list:
            module.p = exponential_dropout_scheduler(p, self.step_num, self.T)


def exponential_dropout_scheduler(dropout_rate, step, max_step):
    return dropout_rate * (1 - np.exp(-10 * step / max_step))


def exponential_scheduler(init_value, end_value, step, max_step):
    step = min(step, max_step - 1)
    return init_value - (init_value - end_value) * (1 - np.exp(-10 * step / max_step))



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-

import json
import logging
import pickle
import re
from pathlib import Path
from typing import Dict, List, Tuple

import numpy as np
import torch
import yaml

from dan.ocr.decoder import CTCLanguageDecoder, GlobalHTADecoder
from dan.ocr.encoder import FCN_Encoder
from dan.ocr.predict.attention import (
    Level,
    get_predicted_polygons_with_confidence,
    parse_delimiters,
    plot_attention,
    split_text_and_confidences,
)
from dan.ocr.transforms import get_preprocessing_transforms
from dan.ocr.utils import load_font
from dan.utils import (
    EntityType,
    ind_to_token,
    list_to_batches,
    pad_images,
    parse_charset_pattern,
    parse_tokens_pattern,
    read_image,
)

logger = logging.getLogger(__name__)


class DAN:
    """
    The DAN class is used to apply a DAN model.
    The class initializes useful parameters: the device and the temperature scalar parameter.
    """

    def __init__(self, device: str, temperature=1.0) -> None:
        """
        Constructor of the DAN class.
        :param device: The device to use.
        """
        super(DAN, self).__init__()
        self.device = device
        self.temperature = temperature

    def load(
        self,
        path: Path,
        mode: str = "eval",
        use_language_model: bool = False,
        compile_model: bool = False,
        dynamic_mode: bool = False,
    ) -> None:
        """
        Load a trained model.
        :param path: Path to the directory containing the model, the YAML parameters file and the charset file.
        :param mode: The mode to load the model (train or eval).
        :param use_language_model: Whether to use an explicit language model to rescore text hypotheses.
        :param compile_model: Whether to compile the model.
        :param dynamic_mode: Whether to use the dynamic mode during model compilation.
        """
        model_path = path / "model.pt"
        assert model_path.is_file(), f"File {model_path} not found"

        params_path = path / "parameters.yml"
        assert params_path.is_file(), f"File {params_path} not found"

        charset_path = path / "charset.pkl"
        assert charset_path.is_file(), f"File {charset_path} not found"

        parameters = yaml.safe_load(params_path.read_text())["parameters"]
        parameters["decoder"]["device"] = self.device

        self.charset = sorted(pickle.loads(charset_path.read_bytes()))

        # Restore the model weights.
        checkpoint = torch.load(model_path, map_location=self.device)

        encoder = FCN_Encoder(parameters["encoder"]).to(self.device)
        encoder.load_state_dict(checkpoint["encoder_state_dict"], strict=True)

        decoder = GlobalHTADecoder(parameters["decoder"]).to(self.device)
        decoder.load_state_dict(checkpoint["decoder_state_dict"], strict=True)

        logger.debug(f"Loaded model {model_path}")

        if compile_model:
            torch.compiler.cudagraph_mark_step_begin()
            encoder = torch.compile(encoder, dynamic=True if dynamic_mode else None)

            torch.compiler.cudagraph_mark_step_begin()
            decoder = torch.compile(decoder, dynamic=True if dynamic_mode else None)

            logger.info("Encoder and decoder have been compiled")

        if mode == "train":
            encoder.train()
            decoder.train()
        elif mode == "eval":
            encoder.eval()
            decoder.eval()
        else:
            raise Exception("Unsupported mode")

        self.encoder = encoder
        self.decoder = decoder
        self.lm_decoder = None

        if use_language_model and parameters["language_model"]["weight"] > 0:
            logger.info(
                f"Decoding with a language model (weight={parameters['language_model']['weight']})."
            )
            self.lm_decoder = CTCLanguageDecoder(
                language_model_path=parameters["language_model"]["model"],
                lexicon_path=parameters["language_model"]["lexicon"],
                tokens_path=parameters["language_model"]["tokens"],
                language_model_weight=parameters["language_model"]["weight"],
            )

        self.mean, self.std = (
            torch.tensor(parameters["mean"]) / 255 if "mean" in parameters else None,
            torch.tensor(parameters["std"]) / 255 if "std" in parameters else None,
        )
        self.preprocessing_transforms = get_preprocessing_transforms(
            parameters.get("preprocessings", [])
        )
        self.max_chars = parameters["max_char_prediction"]

    @property
    def use_lm(self) -> bool:
        """
        Whether the model decodes with a Language Model
        """
        return self.lm_decoder is not None

    def preprocess(self, path: str) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Preprocess an image.
        :param path: Path of the image to load and preprocess.
        """
        image = read_image(path)
        preprocessed_image = self.preprocessing_transforms(image)

        if self.mean is None and self.std is None:
            return preprocessed_image, preprocessed_image

        size = preprocessed_image.shape
        normalized_image = torch.zeros(size)

        mean = self.mean if self.mean is not None else torch.zeros(size[0])
        std = self.std if self.std is not None else torch.ones(size[0])

        for ch in range(size[0]):
            normalized_image[ch, :, :] = (
                preprocessed_image[ch, :, :] - mean[ch]
            ) / std[ch]

        return preprocessed_image, normalized_image

    def predict(
        self,
        input_tensor: torch.Tensor,
        input_sizes: List[torch.Size],
        char_separators: re.Pattern,
        confidences: bool = False,
        attentions: bool = False,
        attention_level: Level = Level.Line,
        extract_objects: bool = False,
        word_separators: re.Pattern = parse_delimiters(["\n", " "]),
        line_separators: re.Pattern = parse_delimiters(["\n"]),
        tokens_separators: re.Pattern | None = None,
        start_token: str | None = None,
        max_object_height: int = 50,
    ) -> dict:
        """
        Run prediction on an input image.
        :param input_tensor: A batch of images to predict.
        :param input_sizes: The original images sizes.
        :param char_separators: The regular expression pattern to split characters.
        :param confidences: Return the characters probabilities.
        :param attentions: Return characters attention weights.
        :param attention_level: Level of text pieces (must be in [char, word, line, ner])
        :param extract_objects: Whether to extract polygons' coordinates.
        :param word_separators: The regular expression pattern to split words.
        :param line_separators: The regular expression pattern to split lines.
        :param tokens_separators: The regular expression pattern to split NER tokens.
        :param start_token: The starting token for the prediction.
        :param max_object_height: Maximum height of predicted objects.
        """
        input_tensor = input_tensor.to(self.device)

        start_token = (
            self.charset.index(start_token) if start_token else len(self.charset) + 1
        )
        end_token = len(self.charset)

        # Run the prediction.
        with torch.no_grad():
            batch_size = input_tensor.size(0)
            reached_end = torch.zeros(
                (batch_size,), dtype=torch.bool, device=self.device
            )
            prediction_len = torch.zeros(
                (batch_size,), dtype=torch.int, device=self.device
            )
            predicted_tokens = (
                torch.ones((batch_size, 1), dtype=torch.long, device=self.device)
                * start_token
            )
            predicted_tokens_len = torch.ones(
                (batch_size,), dtype=torch.int, device=self.device
            )

            # end token index will be used for ctc
            tot_pred = torch.zeros(
                (batch_size, len(self.charset) + 1, self.max_chars),
                dtype=torch.float,
                device=self.device,
            )

            whole_output = list()
            confidence_scores = list()
            attention_maps = list()
            cache = None
            hidden_predict = None

            features = self.encoder(input_tensor.float())
            features_size = features.size()
            features = self.decoder.features_updater.get_pos_features(features)
            features = torch.flatten(features, start_dim=2, end_dim=3).permute(2, 0, 1)

            for i in range(0, self.max_chars):
                (
                    output,
                    pred,
                    hidden_predict,
                    cache,
                    weights,
                ) = self.decoder(
                    features,
                    predicted_tokens,
                    input_sizes,
                    predicted_tokens_len.tolist(),
                    features_size,
                    start=0,
                    hidden_predict=hidden_predict,
                    cache=cache,
                    num_pred=1,
                )

                # output total logit prediction
                tot_pred[:, :, i : i + 1] = pred

                pred = pred / self.temperature
                whole_output.append(output)
                attention_maps.append(weights)
                confidence_scores.append(
                    torch.max(torch.softmax(pred[:, :], dim=1), dim=1).values
                )
                predicted_tokens = torch.cat(
                    [
                        predicted_tokens,
                        torch.argmax(pred[:, :, -1], dim=1, keepdim=True),
                    ],
                    dim=1,
                )
                reached_end = torch.logical_or(
                    reached_end, torch.eq(predicted_tokens[:, -1], end_token)
                )
                predicted_tokens_len += 1

                prediction_len[reached_end == False] = i + 1  # noqa E712

                if torch.all(reached_end):
                    break

            # Concatenate tensors for each token
            confidence_scores = (
                torch.cat(confidence_scores, dim=1).cpu().detach().numpy()
            )
            attention_maps = torch.cat(attention_maps, dim=1).cpu().detach().numpy()

            # Remove bot and eot tokens
            predicted_tokens = predicted_tokens[:, 1:]
            prediction_len[torch.eq(reached_end, False)] = self.max_chars - 1
            predicted_tokens = [
                predicted_tokens[i, : prediction_len[i]] for i in range(batch_size)
            ]
            confidence_scores = [
                confidence_scores[i, : prediction_len[i]].tolist()
                for i in range(batch_size)
            ]

            # Transform tokens to characters
            predicted_text = [
                ind_to_token(self.charset, t, oov_symbol="") for t in predicted_tokens
            ]

            logger.info("Images processed")

        out = {}

        out["text"] = predicted_text
        if self.use_lm:
            out["language_model"] = self.lm_decoder(tot_pred, prediction_len)
        if confidences:
            out["confidences"] = confidence_scores
        if attentions:
            out["attentions"] = attention_maps
        if extract_objects:
            out["objects"] = [
                get_predicted_polygons_with_confidence(
                    predicted_text[i],
                    attention_maps[i],
                    confidence_scores[i],
                    attention_level,
                    input_sizes[i][0],
                    input_sizes[i][1],
                    max_object_height=max_object_height,
                    char_separators=char_separators,
                    word_separators=word_separators,
                    line_separators=line_separators,
                    tokens_separators=tokens_separators,
                )
                for i in range(batch_size)
            ]
        return out


def process_batch(
    image_batch: List[Path],
    dan_model: DAN,
    device: str,
    output: Path,
    confidence_score: bool,
    confidence_score_levels: List[Level],
    attention_map: bool,
    attention_map_level: Level,
    attention_map_scale: float,
    alpha_factor: int | float,
    color_map: str,
    word_separators: List[str],
    line_separators: List[str],
    predict_objects: bool,
    max_object_height: int,
    tokens: Dict[str, EntityType],
    start_token: str,
    font: Path | None = None,
    maximum_font_size: int | None = None,
) -> None:
    input_images, visu_images, input_sizes = [], [], []
    logger.info("Loading images...")
    for image_path in image_batch:
        # Load image and pre-process it
        visu_image, input_image = dan_model.preprocess(str(image_path))
        input_images.append(input_image)
        visu_images.append(visu_image)
        input_sizes.append(input_image.shape[1:])

    # Convert to tensor of size (batch_size, channel, height, width) with batch_size=1
    input_tensor = pad_images(input_images).to(device)
    visu_tensor = pad_images(visu_images).to(device)
    logger.info("Images preprocessed!")

    # Parse delimiters to regex
    char_separators = parse_charset_pattern(dan_model.charset)
    word_separators = parse_delimiters(word_separators)
    line_separators = parse_delimiters(line_separators)

    # NER Entities separators
    ner_separators = parse_tokens_pattern(tokens.values()) if tokens else None
    # Predict
    logger.info("Predicting...")
    prediction = dan_model.predict(
        input_tensor,
        input_sizes,
        confidences=confidence_score,
        attentions=attention_map,
        attention_level=attention_map_level,
        extract_objects=predict_objects,
        char_separators=char_separators,
        word_separators=word_separators,
        line_separators=line_separators,
        tokens_separators=ner_separators,
        max_object_height=max_object_height,
        start_token=start_token,
    )

    logger.info("Prediction parsing...")
    for idx, image_path in enumerate(image_batch):
        predicted_text = prediction["text"][idx]
        result = {"text": predicted_text, "confidences": {}, "language_model": {}}

        if predicted_text:
            # Return LM results
            if dan_model.use_lm:
                result["language_model"] = {
                    "text": prediction["language_model"]["text"][idx],
                    "confidence": prediction["language_model"]["confidence"][idx],
                }

            # Return extracted objects (coordinates, text, confidence)
            if predict_objects:
                result["objects"] = prediction["objects"][idx]

            # Return mean confidence score
            if confidence_score:
                char_confidences = prediction["confidences"][idx]
                result["confidences"]["total"] = np.around(np.mean(char_confidences), 2)

                for level in confidence_score_levels:
                    result["confidences"][level.value] = []
                    texts, confidences, _ = split_text_and_confidences(
                        predicted_text,
                        char_confidences,
                        level,
                        char_separators,
                        word_separators,
                        line_separators,
                        ner_separators,
                    )

                    for text, conf in zip(texts, confidences):
                        result["confidences"][level.value].append(
                            {"text": text, "confidence": conf}
                        )

            # Save gif with attention map
            if attention_map and font and maximum_font_size:
                attentions = prediction["attentions"][idx]
                gif_filename = (
                    f"{output}/{image_path.stem}_{attention_map_level.value}.gif"
                )
                logger.info(f"Creating attention GIF in {gif_filename}")
                plot_attention(
                    image=visu_tensor[idx],
                    text=predicted_text,
                    weights=attentions,
                    level=attention_map_level,
                    scale=attention_map_scale,
                    alpha_factor=alpha_factor,
                    color_map=color_map,
                    char_separators=char_separators,
                    word_separators=word_separators,
                    line_separators=line_separators,
                    tokens_separators=ner_separators,
                    display_polygons=predict_objects,
                    max_object_height=max_object_height,
                    outname=gif_filename,
                    font=font,
                    maximum_font_size=maximum_font_size,
                )
                result["attention_gif"] = gif_filename

        json_filename = Path(output, f"{image_path.stem}.json")
        logger.info(f"Saving JSON prediction in {json_filename}")
        json_filename.write_text(json.dumps(result, indent=2))


def run(
    image_dir: Path,
    model: Path,
    font: Path,
    maximum_font_size: int,
    output: Path,
    confidence_score: bool,
    confidence_score_levels: List[Level],
    attention_map: bool,
    attention_map_level: Level,
    attention_map_scale: float,
    alpha_factor: int | float,
    color_map: str,
    word_separators: List[str],
    line_separators: List[str],
    temperature: float,
    predict_objects: bool,
    max_object_height: int,
    image_extension: str,
    gpu_device: int,
    batch_size: int,
    tokens: Dict[str, EntityType],
    start_token: str,
    use_language_model: bool,
    compile_model: bool,
    dynamic_mode: bool,
) -> None:
    """
    Predict a single image save the output
    :param image_dir: Path to the folder where the images to predict are stored.
    :param model: Path to the directory containing the model, the YAML parameters file and the charset file to use for prediction.
    :param font: Path to the font file to use for the GIF of the attention map.
    :param maximum_font_size: Maximum font size to use for the GIF of the attention map.
    :param output: Path to the output folder where the results will be saved.
    :param confidence_score: Whether to compute confidence score.
    :param confidence_score_levels: Levels of objects to extract.
    :param attention_map: Whether to plot the attention map.
    :param attention_map_level: Level of objects to extract.
    :param attention_map_scale: Scaling factor for the attention map.
    :param alpha_factor: Alpha factor for the attention map.
    :param color_map: A matplotlib colormap to use for the attention maps.
    :param word_separators: List of word separators.
    :param line_separators: List of line separators.
    :param temperature: Temperature scalar parameter.
    :param predict_objects: Whether to extract objects.
    :param max_object_height: Maximum height of predicted objects.
    :param image_extension: Extension of the images to predict.
    :param gpu_device: Use a specific GPU if available.
    :param batch_size: Size of the batches for prediction.
    :param tokens: NER tokens used.
    :param start_token: Use a specific starting token at the beginning of the prediction. Useful when making predictions on different single pages.
    :param use_language_model: Whether to use an explicit language model to rescore text hypotheses.
    :param compile_model: Whether to compile the model.
    :param dynamic_mode: Whether to use the dynamic mode during model compilation.
    """
    # Create output directory if necessary
    if not output.exists():
        output.mkdir()

    # Load model
    cuda_device = f":{gpu_device}" if gpu_device is not None else ""
    device = f"cuda{cuda_device}" if torch.cuda.is_available() else "cpu"
    dan_model = DAN(device, temperature)
    dan_model.load(
        model,
        mode="eval",
        use_language_model=use_language_model,
        compile_model=compile_model,
        dynamic_mode=dynamic_mode,
    )

    # Load font if the attention map is drawn
    if attention_map:
        try:
            load_font(font, maximum_font_size)
        except OSError:
            raise FileNotFoundError(f"The font file is missing at path `{str(font)}`")

    images = image_dir.rglob(f"*{image_extension}")
    for image_batch in list_to_batches(images, n=batch_size):
        process_batch(
            image_batch,
            dan_model,
            device,
            output,
            confidence_score,
            confidence_score_levels,
            attention_map,
            attention_map_level,
            attention_map_scale,
            alpha_factor,
            color_map,
            word_separators,
            line_separators,
            predict_objects,
            max_object_height,
            tokens,
            start_token,
            font,
            maximum_font_size,
        )



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-


class ProcessingError(Exception):
    ...


class ElementProcessingError(ProcessingError):
    """
    Raised when a problem is encountered while processing an element
    """

    element_id: str
    """
    ID of the element being processed.
    """

    def __init__(self, element_id: str, *args: object) -> None:
        super().__init__(*args)
        self.element_id = element_id


class NoTranscriptionError(ElementProcessingError):
    """
    Raised when there are no transcriptions on an element
    """

    def __str__(self) -> str:
        return f"No transcriptions found on element ({self.element_id}) with this config. Skipping."



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-

import json
import shutil
from pathlib import Path

import pytest
import yaml
from prettytable import PrettyTable

from dan.ocr import evaluate
from dan.ocr.manager.metrics import Inference
from dan.ocr.utils import add_metrics_table_row, create_metrics_table
from tests import FIXTURES

PREDICTION_DATA_PATH = FIXTURES / "prediction"


def test_create_metrics_table():
    metric_names = ["ignored", "wer", "cer", "time", "ner"]
    metrics_table = create_metrics_table(metric_names)

    assert isinstance(metrics_table, PrettyTable)
    assert metrics_table.field_names == [
        "Split",
        "CER (HTR-NER)",
        "WER (HTR-NER)",
        "NER",
    ]


def test_add_metrics_table_row():
    metric_names = ["ignored", "wer", "cer", "time", "ner"]
    metrics_table = create_metrics_table(metric_names)

    metrics = {
        "ignored": "whatever",
        "wer": 1.0,
        "cer": 1.3023,
        "time": 42,
    }
    add_metrics_table_row(metrics_table, "train", metrics)

    assert isinstance(metrics_table, PrettyTable)
    assert metrics_table.field_names == [
        "Split",
        "CER (HTR-NER)",
        "WER (HTR-NER)",
        "NER",
    ]
    assert metrics_table.rows == [["train", 130.23, 100, "−"]]


def test_print_worst_predictions(capsys):
    evaluate.print_worst_predictions(
        all_inferences={
            "test": [
                Inference(
                    image="0a56e8b3-95cd-4fa5-a17b-5b0ff9e6ea84.png",
                    ground_truth="ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier Ⓟ12241",
                    prediction="ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241",
                    lm_prediction="ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241",
                    wer=0.125,
                ),
                # Test with empty strings
                Inference(
                    image="0dfe8bcd-ed0b-453e-bf19-cc697012296e.png",
                    ground_truth="Some text",
                    prediction="",
                    lm_prediction="",
                    wer=1,
                ),
                Inference(
                    image="2c242f5c-e979-43c4-b6f2-a6d4815b651d.png",
                    ground_truth="",
                    prediction="Some text",
                    lm_prediction="",
                    wer=1,
                ),
                Inference(
                    image="ffdec445-7f14-4f5f-be44-68d0844d0df1.png",
                    ground_truth="",
                    prediction="",
                    lm_prediction="Some text",
                    wer=0,
                ),
            ]
        }
    )

    # Check the metrics Markdown table
    captured_std = capsys.readouterr()
    last_printed_lines = captured_std.out.split("\n")
    assert (
        "\n".join(last_printed_lines)
        == Path(FIXTURES / "evaluate" / "worst_predictions.md").read_text()
    )


def test_eval_nerval(capsys, evaluate_config):
    evaluate.eval_nerval(
        all_inferences={
            "test": [
                Inference(
                    image="0a56e8b3-95cd-4fa5-a17b-5b0ff9e6ea84.png",
                    ground_truth="ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier Ⓟ12241",
                    prediction="ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241",
                    lm_prediction="ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241",
                    wer=0.125,
                ),
                # Test with empty strings
                Inference(
                    image="0dfe8bcd-ed0b-453e-bf19-cc697012296e.png",
                    ground_truth="Some text",
                    prediction="",
                    lm_prediction="",
                    wer=1,
                ),
                Inference(
                    image="2c242f5c-e979-43c4-b6f2-a6d4815b651d.png",
                    ground_truth="",
                    prediction="Some text",
                    lm_prediction="",
                    wer=1,
                ),
                Inference(
                    image="ffdec445-7f14-4f5f-be44-68d0844d0df1.png",
                    ground_truth="",
                    prediction="",
                    lm_prediction="Some text",
                    wer=0,
                ),
            ]
        },
        tokens=evaluate_config["dataset"]["tokens"],
        threshold=evaluate.NERVAL_THRESHOLD,
    )

    # Check the metrics Markdown table
    captured_std = capsys.readouterr()
    last_printed_lines = captured_std.out.split("\n")
    assert (
        "\n".join(last_printed_lines)
        == Path(FIXTURES / "evaluate" / "eval_nerval.md").read_text()
    )


@pytest.mark.parametrize(
    "training_res, val_res, test_res",
    (
        (
            {
                "nb_chars": 90,
                "cer": 0.1889,
                "nb_chars_no_token": 76,
                "cer_no_token": 0.2105,
                "nb_words": 15,
                "wer": 0.2667,
                "nb_words_no_punct": 15,
                "wer_no_punct": 0.2667,
                "nb_words_no_token": 15,
                "wer_no_token": 0.2667,
                "nb_tokens": 14,
                "ner": 0.0714,
                "nb_samples": 2,
            },
            {
                "nb_chars": 34,
                "cer": 0.0882,
                "nb_chars_no_token": 26,
                "cer_no_token": 0.1154,
                "nb_words": 8,
                "wer": 0.5,
                "nb_words_no_punct": 8,
                "wer_no_punct": 0.5,
                "nb_words_no_token": 8,
                "wer_no_token": 0.5,
                "nb_tokens": 8,
                "ner": 0.0,
                "nb_samples": 1,
            },
            {
                "nb_chars": 36,
                "cer": 0.0278,
                "nb_chars_no_token": 30,
                "cer_no_token": 0.0333,
                "nb_words": 7,
                "wer": 0.1429,
                "nb_words_no_punct": 7,
                "wer_no_punct": 0.1429,
                "nb_words_no_token": 7,
                "wer_no_token": 0.1429,
                "nb_tokens": 6,
                "ner": 0.0,
                "nb_samples": 1,
            },
        ),
    ),
)
@pytest.mark.parametrize("is_output_json", ((True, False)))
def test_evaluate(
    capsys, training_res, val_res, test_res, is_output_json, evaluate_config, tmp_path
):
    evaluate_path = FIXTURES / "evaluate"

    # Use the tmp_path as base folder
    evaluate_config["training"]["output_folder"] = evaluate_path

    output_json = tmp_path / "inference.json" if is_output_json else None

    evaluate_config["training"]["validation"]["font"] = "fonts/LinuxLibertine.ttf"
    evaluate_config["training"]["validation"]["maximum_font_size"] = 32
    evaluate_config["training"]["validation"]["nb_logged_images"] = 5

    evaluate.run(evaluate_config, evaluate.NERVAL_THRESHOLD, output_json=output_json)

    if is_output_json:
        assert json.loads(output_json.read_text()) == json.loads(
            (evaluate_path / "inference.json").read_text()
        )

    # Check that the evaluation results are correct
    for split_name, expected_res in zip(
        ["train", "val", "test"], [training_res, val_res, test_res]
    ):
        filename = (
            evaluate_config["training"]["output_folder"]
            / "results"
            / f"predict_training-{split_name}_1685.yaml"
        )

        assert {
            metric: value
            for metric, value in yaml.safe_load(filename.read_bytes()).items()
            # Remove the times from the results as they vary
            if "time" not in metric
        } == expected_res

    # Remove results files
    shutil.rmtree(evaluate_config["training"]["output_folder"] / "results")

    # Check the metrics Markdown table
    captured_std = capsys.readouterr()
    last_printed_lines = captured_std.out.split("\n")[10:]
    assert (
        "\n".join(last_printed_lines)
        == Path(FIXTURES / "evaluate" / "metrics_table.md").read_text()
    )


@pytest.mark.parametrize(
    "language_model_weight, expected_inferences",
    (
        (
            0.0,
            [
                (
                    "0a56e8b3-95cd-4fa5-a17b-5b0ff9e6ea84.png",  # Image
                    "ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier Ⓟ12241",  # Ground truth
                    "ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241",  # Prediction
                    "",  # LM prediction
                    0.125,  # WER
                ),
                (
                    "0dfe8bcd-ed0b-453e-bf19-cc697012296e.png",  # Image
                    "ⓈTemplié ⒻMarcelle Ⓑ93 ⓁJ Ⓚch ⓄE dachyle",  # Ground truth
                    "ⓈTemplié ⒻMarcelle Ⓑ93 ⓁS Ⓚch ⓄE dactylo Ⓟ18376",  # Prediction
                    "",  # LM prediction
                    0.4286,  # WER
                ),
                (
                    "2c242f5c-e979-43c4-b6f2-a6d4815b651d.png",  # Image
                    "ⓈA ⒻCharles Ⓑ11 ⓁP ⒸC ⓀF ⓄA Ⓟ14331",  # Ground truth
                    "Ⓢd ⒻCharles Ⓑ11 ⓁP ⒸC ⓀF Ⓞd Ⓟ14 31",  # Prediction
                    "",  # LM prediction
                    0.5,  # WER
                ),
                (
                    "ffdec445-7f14-4f5f-be44-68d0844d0df1.png",  # Image
                    "ⓈNaudin ⒻMarie Ⓑ53 ⓁS ⒸV ⓀBelle mère",  # Ground truth
                    "ⓈNaudin ⒻMarie Ⓑ53 ⓁS Ⓒv ⓀBelle mère",  # Prediction
                    "",  # LM prediction
                    0.1429,  # WER
                ),
            ],
        ),
        (
            1.0,
            [
                (
                    "0a56e8b3-95cd-4fa5-a17b-5b0ff9e6ea84.png",  # Image
                    "ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier Ⓟ12241",  # Ground truth
                    "ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241",  # Prediction
                    "ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241",  # LM prediction
                    0.125,  # WER
                ),
                (
                    "0dfe8bcd-ed0b-453e-bf19-cc697012296e.png",  # Image
                    "ⓈTemplié ⒻMarcelle Ⓑ93 ⓁJ Ⓚch ⓄE dachyle",  # Ground truth
                    "ⓈTemplié ⒻMarcelle Ⓑ93 ⓁS Ⓚch ⓄE dactylo Ⓟ18376",  # Prediction
                    "ⓈTemplié ⒻMarcelle Ⓑ93 ⓁS Ⓚch ⓄE dactylo Ⓟ18376",  # LM prediction
                    0.4286,  # WER
                ),
                (
                    "2c242f5c-e979-43c4-b6f2-a6d4815b651d.png",  # Image
                    "ⓈA ⒻCharles Ⓑ11 ⓁP ⒸC ⓀF ⓄA Ⓟ14331",  # Ground truth
                    "Ⓢd ⒻCharles Ⓑ11 ⓁP ⒸC ⓀF Ⓞd Ⓟ14 31",  # Prediction
                    "Ⓢd ⒻCharles Ⓑ11 ⓁP ⒸC ⓀF Ⓞd Ⓟ14 31",  # LM prediction
                    0.5,  # WER
                ),
                (
                    "ffdec445-7f14-4f5f-be44-68d0844d0df1.png",  # Image
                    "ⓈNaudin ⒻMarie Ⓑ53 ⓁS ⒸV ⓀBelle mère",  # Ground truth
                    "ⓈNaudin ⒻMarie Ⓑ53 ⓁS Ⓒv ⓀBelle mère",  # Prediction
                    "ⓈNaudin ⒻMarie Ⓑ53 ⓁS Ⓒv ⓀBelle mère",  # LM prediction
                    0.1429,  # WER
                ),
            ],
        ),
        (
            2.0,
            [
                (
                    "0a56e8b3-95cd-4fa5-a17b-5b0ff9e6ea84.png",  # Image
                    "ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier Ⓟ12241",  # Ground truth
                    "ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241",  # Prediction
                    "ⓈBellisson ⒻGeorges Ⓑ91 ⓁP ⒸM ⓀCh ⓄPlombier ⓅPatron?12241",  # LM prediction
                    0.125,  # WER
                ),
                (
                    "0dfe8bcd-ed0b-453e-bf19-cc697012296e.png",  # Image
                    "ⓈTemplié ⒻMarcelle Ⓑ93 ⓁJ Ⓚch ⓄE dachyle",  # Ground truth
                    "ⓈTemplié ⒻMarcelle Ⓑ93 ⓁS Ⓚch ⓄE dactylo Ⓟ18376",  # Prediction
                    "ⓈTemplié ⒻMarcelle Ⓑ93 ⓁS Ⓚch ⓄE dactylo Ⓟ18376",  # LM prediction
                    0.4286,  # WER
                ),
                (
                    "2c242f5c-e979-43c4-b6f2-a6d4815b651d.png",  # Image
                    "ⓈA ⒻCharles Ⓑ11 ⓁP ⒸC ⓀF ⓄA Ⓟ14331",  # Ground truth
                    "Ⓢd ⒻCharles Ⓑ11 ⓁP ⒸC ⓀF Ⓞd Ⓟ14 31",  # Prediction
                    "Ⓢd ⒻCharles Ⓑ11 ⓁP ⒸC ⓀF Ⓞd Ⓟ14331",  # LM prediction
                    0.5,  # WER
                ),
                (
                    "ffdec445-7f14-4f5f-be44-68d0844d0df1.png",  # Image
                    "ⓈNaudin ⒻMarie Ⓑ53 ⓁS ⒸV ⓀBelle mère",  # Ground truth
                    "ⓈNaudin ⒻMarie Ⓑ53 ⓁS Ⓒv ⓀBelle mère",  # Prediction
                    "ⓈNaudin ⒻMarie Ⓑ53 ⓁS Ⓒv ⓀBelle mère",  # LM prediction
                    0.1429,  # WER
                ),
            ],
        ),
    ),
)
def test_evaluate_language_model(
    capsys, evaluate_config, language_model_weight, expected_inferences, monkeypatch
):
    # LM predictions are never used/displayed
    # We mock the `Inference` class to temporary check the results
    global nb_inferences
    nb_inferences = 0

    class MockInference(Inference):
        def __new__(cls, *args, **kwargs):
            global nb_inferences
            assert args == expected_inferences[nb_inferences]
            nb_inferences += 1

            return super().__new__(cls, *args, **kwargs)

    monkeypatch.setattr("dan.ocr.manager.training.Inference", MockInference)

    # Use the tmp_path as base folder
    evaluate_config["training"]["output_folder"] = FIXTURES / "evaluate"

    # Use a LM decoder
    evaluate_config["model"]["lm"] = {
        "path": PREDICTION_DATA_PATH / "language_model.arpa",
        "weight": language_model_weight,
    }

    evaluate_config["training"]["validation"]["font"] = "fonts/LinuxLibertine.ttf"
    evaluate_config["training"]["validation"]["maximum_font_size"] = 32
    evaluate_config["training"]["validation"]["nb_logged_images"] = 5

    evaluate.run(evaluate_config, evaluate.NERVAL_THRESHOLD, output_json=None)

    # Check that the evaluation results are correct
    for split_name, expected_res in [
        (
            "train",
            {
                "nb_chars": 90,
                "cer": 0.1889,
                "nb_chars_no_token": 76,
                "cer_no_token": 0.2105,
                "nb_words": 15,
                "wer": 0.2667,
                "nb_words_no_punct": 15,
                "wer_no_punct": 0.2667,
                "nb_words_no_token": 15,
                "wer_no_token": 0.2667,
                "nb_tokens": 14,
                "ner": 0.0714,
                "nb_samples": 2,
            },
        ),
        (
            "val",
            {
                "nb_chars": 34,
                "cer": 0.0882,
                "nb_chars_no_token": 26,
                "cer_no_token": 0.1154,
                "nb_words": 8,
                "wer": 0.5,
                "nb_words_no_punct": 8,
                "wer_no_punct": 0.5,
                "nb_words_no_token": 8,
                "wer_no_token": 0.5,
                "nb_tokens": 8,
                "ner": 0.0,
                "nb_samples": 1,
            },
        ),
        (
            "test",
            {
                "nb_chars": 36,
                "cer": 0.0278,
                "nb_chars_no_token": 30,
                "cer_no_token": 0.0333,
                "nb_words": 7,
                "wer": 0.1429,
                "nb_words_no_punct": 7,
                "wer_no_punct": 0.1429,
                "nb_words_no_token": 7,
                "wer_no_token": 0.1429,
                "nb_tokens": 6,
                "ner": 0.0,
                "nb_samples": 1,
            },
        ),
    ]:
        filename = (
            evaluate_config["training"]["output_folder"]
            / "results"
            / f"predict_training-{split_name}_1685.yaml"
        )

        with filename.open() as f:
            assert {
                metric: value
                for metric, value in yaml.safe_load(f).items()
                # Remove the times from the results as they vary
                if "time" not in metric
            } == expected_res

    # Remove results files
    shutil.rmtree(evaluate_config["training"]["output_folder"] / "results")

    # Check the metrics Markdown table
    captured_std = capsys.readouterr()
    last_printed_lines = captured_std.out.split("\n")[10:]
    assert (
        "\n".join(last_printed_lines)
        == Path(FIXTURES / "evaluate" / "metrics_table.md").read_text()
    )



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
import logging
import re
from dataclasses import dataclass, field
from typing import Dict, List

from lxml.etree import Element, SubElement, tostring

from arkindex_export import TranscriptionEntity
from dan.utils import EntityType

logger = logging.getLogger(__name__)

# Replace \t with regular space and consecutive spaces
TRIM_SPACE_REGEX = re.compile(r"[\t ]+")
TRIM_RETURN_REGEX = re.compile(r"[\r\n]+")

# Remove invalid characters to build valid XML tag name
SLUG_PATTERN = re.compile(r"[\W]+")

# Some characters are encoded in XML but we don't want them encoded in the end
ENCODING_MAP = {
    "&#13;": "\r",
    "&lt;": "<",
    "&gt;": ">",
    "&amp;": "&",
}


def normalize_linebreaks(text: str) -> str:
    """
    Remove begin/ending linebreaks.
    Replace \r with regular linebreak and consecutive linebreaks.
    :param text: Text to normalize.
    """
    return TRIM_RETURN_REGEX.sub("\n", text.strip())


def normalize_spaces(text: str) -> str:
    """
    Remove begin/ending spaces.
    Replace \t with regular space and consecutive spaces.
    :param text: Text to normalize.
    """
    return TRIM_SPACE_REGEX.sub(" ", text.strip())


def slugify(text: str):
    """
    Replace invalid characters in text to underscores to use it as XML tag.
    """
    return SLUG_PATTERN.sub("_", text)


def get_translation_map(tokens: Dict[str, EntityType]) -> Dict[str, str] | None:
    if not tokens:
        return

    translation_map = {
        # Roots
        "<root>": "",
        "</root>": "",
    }
    # Tokens
    for entity_name, token_type in tokens.items():
        translation_map[f"<{slugify(entity_name)}>"] = token_type.start
        translation_map[f"</{slugify(entity_name)}>"] = token_type.end

    return translation_map


@dataclass
class XMLEntity:
    type: str
    name: str
    offset: int
    length: int
    worker_version: str
    worker_run: str
    children: List["XMLEntity"] = field(default_factory=list)

    @property
    def end(self) -> int:
        return self.offset + self.length

    def add_child(self, child: TranscriptionEntity):
        self.children.append(
            XMLEntity(
                type=child["type"],
                name=child["name"],
                offset=child["offset"] - self.offset,
                length=child["length"],
                worker_version=child["worker_version"],
                worker_run=child["worker_run"],
            )
        )

    def insert(self, parent: Element):
        e = SubElement(parent, slugify(self.type))

        if not self.children:
            # No children
            e.text = self.name
            return

        offset = 0
        for child in self.children:
            # Add text before entity
            portion_before = self.name[offset : child.offset]
            offset += len(portion_before)
            if len(e):
                e[-1].tail = portion_before
            else:
                e.text = portion_before
            child.insert(e)
            offset += child.length

        # Text after the last entity
        e[-1].tail = self.name[self.children[-1].end : self.end]


def entities_to_xml(
    text: str,
    predictions: List[TranscriptionEntity],
    entity_separators: List[str] | None = None,
) -> str:
    """Represent the transcription and its entities in XML format. Each entity will be exposed with an XML tag.
    Its type will be used to name the tag.

    :param text: The text of the transcription
    :param predictions: The list of entities linked to the transcription
    :param entity_separators: When provided, instead of adding the text between entities, add one separator encountered in this text. The order is kept when looking for separators. Defaults to None
    :return: The representation of the transcription in XML format
    """

    def _find_separator(transcription: str) -> str:
        """
        Find the first entity separator in the provided transcription.
        """
        for separator in entity_separators:
            if separator in transcription:
                return separator
        return ""

    def add_portion(entity_offset: int | None = None):
        """
        Add the portion of text between entities either:
        - after the last node, if there is one before
        - on this node

        If we remove the text between entities, we keep one of the separators provided. Order matters.
        """
        portion = text[offset:entity_offset]

        if entity_separators:
            # Remove the text except the first entity_separator encountered
            portion = _find_separator(portion)

        if len(root):
            root[-1].tail = portion
        else:
            root.text = portion

    entities = iter(predictions)

    # This will mark the ending position of the first-level of entities
    last_end = None
    parsed: List[XMLEntity] = []

    for entity in entities:
        # First entity is not inside any other
        # If offset is too high, no nestation
        if not last_end or entity["offset"] >= last_end:
            parsed.append(XMLEntity(**entity))
            last_end = entity["offset"] + entity["length"]
            continue

        # Nested entity
        parsed[-1].add_child(entity)

    # XML export
    offset = 0
    root = Element("root")

    for entity in parsed:
        add_portion(entity.offset)

        entity.insert(root)

        offset = entity.end

    # Add text after last entity
    add_portion()

    # Cleanup separators introduced when text was removed
    if entity_separators:
        characters = "".join(entity_separators)
        root.text = root.text.lstrip(characters)
        # Strip trailing spaces on last child
        root[-1].tail = root[-1].tail.rstrip(characters)

    encoded_transcription = tostring(root, encoding="utf-8").decode()
    for pattern, repl in ENCODING_MAP.items():
        encoded_transcription = encoded_transcription.replace(pattern, repl)
    return encoded_transcription



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
import re


def convert(text):
    return int(text) if text.isdigit() else text.lower()


def natural_sort(data):
    return sorted(data, key=lambda key: [convert(c) for c in re.split("([0-9]+)", key)])



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-

import functools
import json
import logging
import pickle
from collections import defaultdict
from concurrent.futures import Future, ThreadPoolExecutor
from itertools import chain
from pathlib import Path
from typing import Dict, List, Tuple

import cv2
import numpy as np
from PIL import Image
from tqdm import tqdm

from dan.datasets.download.exceptions import ImageDownloadError
from dan.datasets.download.utils import (
    download_image,
    get_bbox,
)
from dan.datasets.extract.arkindex import TRAIN_NAME
from line_image_extractor.extractor import extract
from line_image_extractor.image_utils import (
    BoundingBox,
    Extraction,
    polygon_to_bbox,
)

IMAGES_DIR = "images"  # Subpath to the images directory.

IIIF_URL = "{image_url}/{bbox}/{size}/0/default.jpg"
# IIIF 2.0 uses `full`
IIIF_FULL_SIZE = "full"

logger = logging.getLogger(__name__)


def output_attr_required(func):
    """
    Always check that the output attribute is not null.
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        assert self.output is not None, "Define an output folder to download images."
        return func(self, *args, **kwargs)

    return wrapper


class ImageDownloader:
    """
    Download images from extracted data
    """

    def __init__(
        self,
        output: Path | None = None,
        max_width: int | None = None,
        max_height: int | None = None,
        image_extension: str = "",
        unknown_token: str = "⁇",
    ) -> None:
        self.output: Path | None = output

        self.max_width: int | None = max_width
        self.max_height: int | None = max_height
        self.image_extension: str = image_extension
        self.data: dict[str, dict] = defaultdict(dict)
        self.unknown_token: str = unknown_token

    @output_attr_required
    def load_split_data(self) -> None:
        """
        Load the dataset stored in `split.json` and initializes the charset.
        """
        # Load split file
        split_file = self.output / "split.json"
        self.split: Dict = (
            json.loads(split_file.read_text())
            if split_file and split_file.is_file()
            else {}
        )

        # Create directories
        for split_name in self.split:
            (self.output / IMAGES_DIR / split_name).mkdir(parents=True, exist_ok=True)

        self.charset = set(
            chain.from_iterable(
                split_data["text"] for split_data in self.split[TRAIN_NAME].values()
            )
        )

        # Add unknown token to charset
        self.charset.add(self.unknown_token)

    def check_extraction(self, values: dict) -> str | None:
        # Check dataset_id parameter
        if values.get("dataset_id") is None:
            return "Dataset ID not found"

        # Check image parameters
        if not (image := values.get("image")):
            return "Image information not found"

        # Only support iiif_url with polygon for now
        if not image.get("iiif_url"):
            return "Image IIIF URL not found"
        if not image.get("polygon"):
            return "Image polygon not found"

        # Check text parameter
        if values.get("text") is None:
            return "Text not found"

        if self.unknown_token in values["text"]:
            return "Unknown token found in the transcription text"

    def get_iiif_size_arg(self, width: int, height: int) -> str:
        if (self.max_width is None or width <= self.max_width) and (
            self.max_height is None or height <= self.max_height
        ):
            return IIIF_FULL_SIZE

        bigger_width = self.max_width and width >= self.max_width
        bigger_height = self.max_height and height >= self.max_height

        if bigger_width and bigger_height:
            # Resize to the biggest dim to keep aspect ratio
            # Only resize width is bigger than max size
            # This ratio tells which dim needs the biggest shrinking
            ratio = width * self.max_height / (height * self.max_width)
            return f"{self.max_width}," if ratio > 1 else f",{self.max_height}"
        elif bigger_width:
            return f"{self.max_width},"
        # Only resize height is bigger than max size
        elif bigger_height:
            return f",{self.max_height}"

    def build_iiif_url(
        self, polygon: List[List[int]], image_url: str
    ) -> Tuple[BoundingBox, str]:
        bbox = polygon_to_bbox(polygon)
        size = self.get_iiif_size_arg(width=bbox.width, height=bbox.height)
        # Rotations are done using the lib
        return IIIF_URL.format(image_url=image_url, bbox=get_bbox(polygon), size=size)

    @output_attr_required
    def build_tasks(self) -> List[Dict[str, str]]:
        tasks = []
        for split, items in self.split.items():
            # Create directories
            destination = self.output / IMAGES_DIR / split
            destination.mkdir(parents=True, exist_ok=True)

            for element_id, values in items.items():
                filename = Path(element_id).with_suffix(self.image_extension)

                error = self.check_extraction(values)
                if error:
                    logger.warning(f"{destination / filename}: {error}")
                    continue

                image_path = destination / values["dataset_id"] / filename
                image_path.parent.mkdir(parents=True, exist_ok=True)

                # Replace unknown characters by the unknown token
                if split != TRAIN_NAME:
                    unknown_charset = set(values["text"]) - self.charset
                    values["text"] = values["text"].translate(
                        {
                            ord(unknown_char): self.unknown_token
                            for unknown_char in unknown_charset
                        }
                    )

                # Store a relative path to the label file in case we need to move the data elsewhere
                self.data[split][str(image_path.relative_to(self.output))] = values[
                    "text"
                ]

                # Create task for multithreading pool if image does not exist yet
                if image_path.exists():
                    continue

                polygon = values["image"]["polygon"]
                iiif_url = values["image"]["iiif_url"]
                tasks.append(
                    {
                        "split": split,
                        "polygon": polygon,
                        "image_url": self.build_iiif_url(polygon, iiif_url),
                        "destination": image_path,
                    }
                )
        return tasks

    def get_image(
        self,
        split: str,
        polygon: List[List[int]],
        image_url: str,
        destination: Path,
    ) -> None:
        """Save the element's image to the given path and applies any image operations needed.

        :param split: Dataset split this image belongs to.
        :param polygon: Polygon of the processed element.
        :param image_url: Base IIIF URL of the image.
        :param destination: Where the image should be saved.
        """
        bbox = polygon_to_bbox(polygon)
        try:
            img: Image.Image = download_image(image_url)

            # The polygon's coordinate are in the referential of the full image
            # We need to remove the offset of the bounding rectangle
            polygon = [(x - bbox.x, y - bbox.y) for x, y in polygon]

            # Normalize bbox
            bbox = BoundingBox(x=0, y=0, width=bbox.width, height=bbox.height)

            image = extract(
                img=cv2.cvtColor(np.asarray(img), cv2.COLOR_RGB2BGR),
                polygon=np.asarray(polygon).clip(0),
                bbox=bbox,
                extraction_mode=Extraction.boundingRect,
                max_deskew_angle=45,
            )

            cv2.imwrite(str(destination), image)

        except Exception as e:
            raise ImageDownloadError(
                split=split, path=destination, url=image_url, exc=e
            )

    def download_images(self, tasks: List[Dict[str, str]]) -> None:
        """
        Execute each image download task in parallel

        :param tasks: List of tasks to execute.
        """
        failed_downloads = []
        with tqdm(
            desc="Downloading images", total=len(tasks)
        ) as pbar, ThreadPoolExecutor() as executor:

            def process_future(future: Future):
                """
                Callback function called at the end of the thread
                """
                # Update the progress bar count
                pbar.update(1)

                exc = future.exception()
                if exc is None:
                    # No error
                    return
                # If failed, tag for removal
                assert isinstance(exc, ImageDownloadError)
                # Remove transcription from labels dict
                del self.data[exc.split][exc.path]
                # Save tried URL
                failed_downloads.append((exc.url, exc.message))

            # Submit all tasks
            for task in tasks:
                executor.submit(self.get_image, **task).add_done_callback(
                    process_future
                )

        if failed_downloads:
            logger.error(f"Failed to download {len(failed_downloads)} image(s).")
            print(*list(map(": ".join, failed_downloads)), sep="\n")

    @output_attr_required
    def export(self) -> None:
        """
        Writes a `labels.json` file containing a mapping of the images that have been correctly uploaded (identified by its path)
        to the ground-truth transcription (with NER tokens if needed).
        """
        (self.output / "labels.json").write_text(
            json.dumps(
                self.data,
                sort_keys=True,
                indent=4,
            )
        )

        (self.output / "charset.pkl").write_bytes(
            pickle.dumps(sorted(list(self.charset)))
        )

    @output_attr_required
    def run(self) -> None:
        """
        Download the missing images from a `split.json` file and build a `labels.json` file containing
        a mapping of the images that have been correctly uploaded (identified by its path)
        to the ground-truth transcription (with NER tokens if needed).
        """
        self.load_split_data()
        tasks: list[dict[str, str]] = self.build_tasks()
        self.download_images(tasks)
        self.export()


def run(
    output: Path,
    max_width: int | None,
    max_height: int | None,
    image_format: str,
    unknown_token: str,
):
    """
    Download the missing images from a `split.json` file and build a `labels.json` file containing
    a mapping of the images that have been correctly uploaded (identified by its path)
    to the ground-truth transcription (with NER tokens if needed).

    :param output: Path where the `split.json` file is stored and where the data will be generated
    :param max_width: Images larger than this width will be resized to this width
    :param max_height: Images larger than this height will be resized to this height
    :param image_format: Images will be saved under this format
    :param unknown_token: The token used to replace unknown characters.
    """
    ImageDownloader(
        output=output,
        max_width=max_width,
        max_height=max_height,
        image_extension=image_format,
        unknown_token=unknown_token,
    ).run()



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
import pickle
import random

import numpy as np
import torch
from torch.utils.data import DataLoader
from torch.utils.data.distributed import DistributedSampler

from dan.ocr.manager.dataset import OCRDataset
from dan.ocr.transforms import get_augmentation_transforms, get_preprocessing_transforms
from dan.utils import pad_images, pad_sequences_1D


class OCRDatasetManager:
    def __init__(
        self, dataset_params: dict, training_params: dict, device: torch.device
    ):
        self.params = dataset_params
        self.training_params = training_params
        self.device_params = training_params["device"]

        # Whether data should be copied on GPU via https://pytorch.org/docs/stable/generated/torch.Tensor.pin_memory.html
        self.pin_memory = device != torch.device("cpu")

        self.train_dataset = None
        self.valid_datasets = dict()
        self.test_datasets = dict()

        self.train_loader = None
        self.valid_loaders = dict()
        self.test_loaders = dict()

        self.train_sampler = None
        self.valid_samplers = dict()
        self.test_samplers = dict()

        self.mean = None
        self.std = None

        self.generator = torch.Generator()
        self.generator.manual_seed(0)

        self.load_in_memory = self.training_params["data"].get("load_in_memory", True)
        self.charset = self.get_charset()
        self.tokens = self.get_tokens()
        self.training_params["data"]["padding_token"] = self.tokens["pad"]

        self.my_collate_function = OCRCollateFunction(
            padding_token=training_params["data"]["padding_token"]
        )
        self.augmentation = (
            get_augmentation_transforms()
            if self.training_params["data"]["augmentation"]
            else None
        )
        self.preprocessing = get_preprocessing_transforms(
            training_params["data"]["preprocessings"], to_pil_image=True
        )

    def load_datasets(self):
        """
        Load training and validation datasets
        """
        self.train_dataset = OCRDataset(
            set_name="train",
            paths_and_sets=self.get_paths_and_sets(self.params["train"]["datasets"]),
            charset=self.charset,
            tokens=self.tokens,
            preprocessing_transforms=self.preprocessing,
            augmentation_transforms=self.augmentation,
            load_in_memory=self.load_in_memory,
            mean=self.mean,
            std=self.std,
        )

        self.mean, self.std = self.train_dataset.compute_std_mean()

        for custom_name in self.params["val"]:
            self.valid_datasets[custom_name] = OCRDataset(
                set_name="val",
                paths_and_sets=self.get_paths_and_sets(self.params["val"][custom_name]),
                charset=self.charset,
                tokens=self.tokens,
                preprocessing_transforms=self.preprocessing,
                augmentation_transforms=None,
                load_in_memory=self.load_in_memory,
                mean=self.mean,
                std=self.std,
            )

    def load_ddp_samplers(self):
        """
        Load training and validation data samplers
        """
        if self.device_params["use_ddp"]:
            self.train_sampler = DistributedSampler(
                self.train_dataset,
                num_replicas=self.device_params["nb_gpu"],
                rank=self.device_params["ddp_rank"],
                shuffle=True,
            )
            for custom_name in self.valid_datasets:
                self.valid_samplers[custom_name] = DistributedSampler(
                    self.valid_datasets[custom_name],
                    num_replicas=self.device_params["nb_gpu"],
                    rank=self.device_params["ddp_rank"],
                    shuffle=False,
                )
        else:
            for custom_name in self.valid_datasets:
                self.valid_samplers[custom_name] = None

    def load_dataloaders(self):
        """
        Load training and validation data loaders
        """
        self.train_loader = DataLoader(
            self.train_dataset,
            batch_size=self.training_params["data"]["batch_size"],
            shuffle=True if self.train_sampler is None else False,
            drop_last=False,
            sampler=self.train_sampler,
            num_workers=self.device_params["nb_gpu"]
            * self.training_params["data"]["worker_per_gpu"],
            pin_memory=self.pin_memory,
            collate_fn=self.my_collate_function,
            worker_init_fn=self.seed_worker,
            generator=self.generator,
        )

        for key in self.valid_datasets:
            self.valid_loaders[key] = DataLoader(
                self.valid_datasets[key],
                batch_size=1,
                sampler=self.valid_samplers[key],
                shuffle=False,
                num_workers=self.device_params["nb_gpu"]
                * self.training_params["data"]["worker_per_gpu"],
                pin_memory=self.pin_memory,
                drop_last=False,
                collate_fn=self.my_collate_function,
                worker_init_fn=self.seed_worker,
                generator=self.generator,
            )

    @staticmethod
    def seed_worker(worker_id):
        """
        Set worker seed
        """
        worker_seed = torch.initial_seed() % 2**32
        np.random.seed(worker_seed)
        random.seed(worker_seed)

    def generate_test_loader(self, custom_name, sets_list):
        """
        Load test dataset, data sampler and data loader
        """
        if custom_name in self.test_loaders:
            return
        paths_and_sets = list()
        for set_info in sets_list:
            paths_and_sets.append(
                {"path": self.params["datasets"][set_info[0]], "set_name": set_info[1]}
            )
        self.test_datasets[custom_name] = OCRDataset(
            set_name="test",
            paths_and_sets=paths_and_sets,
            charset=self.charset,
            tokens=self.tokens,
            preprocessing_transforms=self.preprocessing,
            augmentation_transforms=None,
            load_in_memory=self.load_in_memory,
            mean=self.mean,
            std=self.std,
        )

        if self.device_params["use_ddp"]:
            self.test_samplers[custom_name] = DistributedSampler(
                self.test_datasets[custom_name],
                num_replicas=self.device_params["nb_gpu"],
                rank=self.device_params["ddp_rank"],
                shuffle=False,
            )
        else:
            self.test_samplers[custom_name] = None

        self.test_loaders[custom_name] = DataLoader(
            self.test_datasets[custom_name],
            batch_size=1,
            sampler=self.test_samplers[custom_name],
            shuffle=False,
            num_workers=self.device_params["nb_gpu"]
            * self.training_params["data"]["worker_per_gpu"],
            pin_memory=self.pin_memory,
            drop_last=False,
            collate_fn=self.my_collate_function,
            worker_init_fn=self.seed_worker,
            generator=self.generator,
        )

    def get_paths_and_sets(self, dataset_names_folds):
        """
        Set the right path for each data set
        """
        paths_and_sets = list()
        for dataset_name, fold in dataset_names_folds:
            path = self.params["datasets"][dataset_name]
            paths_and_sets.append({"path": path, "set_name": fold})
        return paths_and_sets

    def get_charset(self):
        """
        Merge the charset of the different datasets used
        """
        if "charset" in self.params:
            return self.params["charset"]
        datasets = self.params["datasets"]
        charset = set()
        for key in datasets:
            charset = charset.union(
                set(pickle.loads((datasets[key] / "charset.pkl").read_bytes()))
            )
        if "" in charset:
            charset.remove("")
        return sorted(list(charset))

    def get_tokens(self):
        """
        Get special tokens
        """
        return {
            "end": len(self.charset),
            "start": len(self.charset) + 1,
            "pad": len(self.charset) + 2,
        }


class OCRCollateFunction:
    """
    Merge samples data to mini-batch data for OCR task
    """

    def __init__(self, padding_token):
        self.label_padding_value = padding_token

    def __call__(self, batch_data):
        formatted_batch_data = {
            "imgs": pad_images(
                [
                    torch.from_numpy(sample["img"]).permute(2, 0, 1)
                    for sample in batch_data
                ]
            ),
            "labels": pad_sequences_1D(
                [sample["token_label"] for sample in batch_data],
                padding_value=self.label_padding_value,
            ).long(),
        }

        formatted_batch_data.update(
            {
                formatted_key: [sample[initial_key] for sample in batch_data]
                for formatted_key, initial_key in zip(
                    [
                        "names",
                        "imgs_position",
                        "imgs_reduced_shape",
                        "labels_len",
                        "raw_labels",
                    ],
                    ["name", "img_position", "img_reduced_shape", "label_len", "label"],
                )
            }
        )
        return formatted_batch_data



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-
import argparse
import errno

from dan.bio import add_convert_bio_parser
from dan.datasets import add_dataset_parser
from dan.ocr import add_evaluate_parser, add_predict_parser, add_train_parser


def get_parser():
    parser = argparse.ArgumentParser(prog="teklia-dan")
    subcommands = parser.add_subparsers(metavar="subcommand")

    add_convert_bio_parser(subcommands)
    add_dataset_parser(subcommands)
    add_train_parser(subcommands)
    add_evaluate_parser(subcommands)
    add_predict_parser(subcommands)
    return parser


def main():
    parser = get_parser()
    args = vars(parser.parse_args())
    if "func" in args:
        # Run the subcommand's function
        try:
            status = args.pop("func")(**args)
            parser.exit(status=status)
        except KeyboardInterrupt:
            # Just quit silently on ^C instead of displaying a long traceback
            parser.exit(status=errno.EOWNERDEAD)
    else:
        parser.error("A subcommand is required.")



# Copyright Teklia (contact@teklia.com) & Denis Coquenet
# This code is licensed under CeCILL-C

# -*- coding: utf-8 -*-

from operator import itemgetter

import pytest

from arkindex_export import Dataset, DatasetElement, Element
from dan.datasets.extract.arkindex import TRAIN_NAME
from dan.datasets.extract.db import (
    get_dataset_elements,
    get_elements,
    get_transcription_entities,
    get_transcriptions,
)


def test_get_dataset_elements(mock_database):
    """
    Assert dataset elements retrieval output against verified results
    """
    dataset_elements = get_dataset_elements(
        dataset=Dataset.select().get(),
        split=TRAIN_NAME,
    )

    # ID verification
    assert all(
        isinstance(dataset_element, DatasetElement)
        for dataset_element in dataset_elements
    )
    assert [dataset_element.element.id for dataset_element in dataset_elements] == [
        "train-page_1",
        "train-page_2",
    ]


def test_get_elements(mock_database):
    """
    Assert elements retrieval output against verified results
    """
    elements = get_elements(
        parent_id="train-page_1",
        element_type=["text_line"],
    )

    # ID verification
    assert all(isinstance(element, Element) for element in elements)
    assert [element.id for element in elements] == [
        "train-page_1-line_1",
        "train-page_1-line_2",
        "train-page_1-line_3",
        "train-page_1-line_4",
    ]


@pytest.mark.parametrize(
    "sources",
    ([False], ["id"], [], [False, "id"]),
)
def test_get_transcriptions(sources, mock_database):
    """
    Assert transcriptions retrieval output against verified results
    """
    worker_versions = [
        f"worker_version_{source}" if isinstance(source, str) else source
        for source in sources
    ]
    worker_runs = [
        f"worker_run_{source}" if isinstance(source, str) else source
        for source in sources
    ]

    element_id = "train-page_1-line_1"
    transcriptions = get_transcriptions(
        element_id=element_id,
        transcription_worker_versions=worker_versions,
        transcription_worker_runs=worker_runs,
    )

    expected_transcriptions = []
    if not sources or False in sources:
        expected_transcriptions.append(
            {
                "text": "Laulont  Francois  8",
                "worker_version": None,
                "worker_run": None,
            }
        )

    if not sources or "id" in sources:
        expected_transcriptions.append(
            {
                "text": "laulont  francois  8",
                "worker_version": "worker_version_id",
                "worker_run": "worker_run_id",
            }
        )

    # Do not compare IDs
    transcriptions = transcriptions.dicts()
    for transcription in transcriptions:
        del transcription["id"]

    assert sorted(transcriptions, key=itemgetter("text")) == expected_transcriptions


@pytest.mark.parametrize("source", (False, "id", None))
@pytest.mark.parametrize(
    "supported_types", (["surname"], ["surname", "firstname", "age"])
)
def test_get_transcription_entities(source, mock_database, supported_types):
    worker_version = f"worker_version_{source}" if isinstance(source, str) else source
    worker_run = f"worker_run_{source}" if isinstance(source, str) else source

    transcription_id = "train-page_1-line_1" + ("source" if source else "")
    entities = get_transcription_entities(
        transcription_id=transcription_id,
        entity_worker_versions=[worker_version] if worker_version is not None else [],
        entity_worker_runs=[worker_run] if worker_run is not None else [],
        supported_types=supported_types,
    )

    expected_entities = [
        {
            "name": "Laulont",
            "type": "surname",
            "offset": 0,
            "length": 7,
        },
        {
            "name": "Francois",
            "type": "firstname",
            "offset": 9,
            "length": 8,
        },
        {
            "name": "8",
            "type": "age",
            "offset": 19,
            "length": 1,
        },
    ]

    expected_entities = list(
        filter(lambda ent: ent["type"] in supported_types, expected_entities)
    )
    for entity in expected_entities:
        if source:
            entity["name"] = entity["name"].lower()
        entity["worker_version"] = worker_version or None
        entity["worker_run"] = worker_run or None

    assert sorted(entities, key=itemgetter("offset")) == expected_entities
